# Copyright (C) 2011 Adam Stein <adam@csh.rit.edu>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

"""Reports/Text Reports/Filtered Descendant Report"""

#------------------------------------------------------------------------
#
# standard python modules
#
#------------------------------------------------------------------------
import os

#------------------------------------------------------------------------
#
# GRAMPS modules
#
#------------------------------------------------------------------------
import DateHandler
from Enhanced_DeepRelationshipPathBetween import find_deep_relations
import Enhanced_Endnotes
from Errors import ReportError
from gen.display.name import displayer as name_displayer
from gen.ggettext import gettext as _
from gen.lib import FamilyRelType, NoteType, Person
from gen.plug.docgen import (FontStyle, IndexMark, ParagraphStyle,
                             FONT_SANS_SERIF, FONT_SERIF, INDEX_TYPE_TOC,
                             PARA_ALIGN_CENTER)
from gen.plug.menu import (BooleanOption, EnumeratedListOption, FilterOption,
                           NumberOption, StringOption, TextOption)
from gen.plug.report import (Report, Bibliography)
from gen.plug.report import utils as ReportUtils
from gui.plug.report import MenuReportOptions
from libnarrate import Narrator
from libsubstkeyword import SubstKeywords
from Utils import media_path_full

#------------------------------------------------------------------------
#
# Constants
#
#------------------------------------------------------------------------
EMPTY_ENTRY = "_____________"
HENRY = "123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ"

#------------------------------------------------------------------------
#
# DescendantReport
#
#------------------------------------------------------------------------
class DescendantReport(Report):
    def __init__(self, database, options_class):
        """
        Create the DescendantReport object that produces the report.
                        
        The arguments are:

        database        - the GRAMPS database instance
        options_class   - instance of the Options class for this report

        The following parameters are defined in the options class:
                                                                        
        calcage         - Whether to compute age.
        caption_fmt     - Caption format.
        desref          - Whether to add descendant references in child list.
        omit_duplicate  - Whether to omit duplicate ancestors (e.g. when distant cousins mary).
        filter          - Filter to be applied to the people of the database.
        honor_private   - Honor the private flag setting when getting information
        inc_addr        - Whether to include addresses.
        inc_attrs       - Whether to include attributes.
        inc_captions    - Whether to include captions under photos.
        inc_events      - Whether to include events.
        inc_mates       - Whether to include information about partners
        inc_names       - Whether to include other names.
        inc_notes       - Whether to include notes.
        inc_paths       - Whether to include path to the start person.
        inc_photos      - Whether to include images.
        inc_sources     - Whether to include source references.
        inc_srcnotes    - Whether to include source notes in the Endnotes section
        inc_succession  - Whether to indicate which child has succession.
        links           - Whether to turn URLs into real links
        listchildren    - Whether to list children.
        max_generations - Maximum number of generations to include.
        pgbrkenotes     - Whether to include page break before End Notes.
        pgbrk           - Whether to include page breaks between generations.
        title           - Report title
        use_call        - Whether to use the call name as the first name.
        use_fulldate    - Whether to use full dates instead of just year.
        verbose         - Whether to use complete sentences.
        """

        Report.__init__(self, database, options_class)

        menu = options_class.menu

        self.calcage         = menu.get_option_by_name("calcage").get_value()
        self.caption_fmt     = menu.get_option_by_name("caption")
        self.desref          = menu.get_option_by_name("desref").get_value()
        self.filter          = menu.get_option_by_name("filter")
        self.honor_private   = menu.get_option_by_name("honor_private").get_value()
        self.inc_addr        = menu.get_option_by_name("inc_addr").get_value()
        self.inc_attrs       = menu.get_option_by_name("inc_attrs").get_value()
        self.inc_captions    = menu.get_option_by_name("inc_captions").get_value()
        self.inc_events      = menu.get_option_by_name("inc_events").get_value()
        self.inc_mates       = menu.get_option_by_name("inc_mates").get_value()
        self.inc_names       = menu.get_option_by_name("inc_names").get_value()
        self.inc_notes       = menu.get_option_by_name("inc_notes").get_value()
        self.inc_paths       = menu.get_option_by_name("inc_paths").get_value()
        self.inc_photos      = menu.get_option_by_name("inc_photos").get_value()
        self.inc_sources     = menu.get_option_by_name("inc_sources").get_value()
        self.inc_srcnotes    = menu.get_option_by_name("inc_srcnotes").get_value()
        self.inc_succession  = menu.get_option_by_name("inc_succession").get_value()
        self.links           = menu.get_option_by_name("links").get_value()
        self.listchildren    = menu.get_option_by_name("listchildren").get_value()
        self.max_generations = menu.get_option_by_name("max_generations").get_value()
        self.omit_duplicate  = menu.get_option_by_name("omit_duplicate").get_value()
        self.pgbrk           = menu.get_option_by_name("pgbrk").get_value()
        self.pgbrkenotes     = menu.get_option_by_name("pgbrkenotes").get_value()
        self.title           = menu.get_option_by_name("title").get_value()
        self.use_fulldate    = menu.get_option_by_name("use_fulldate").get_value()
        self.verbose         = menu.get_option_by_name("verbose").get_value()

        replace_date         = menu.get_option_by_name("replace_date").get_value()
        replace_place        = menu.get_option_by_name("replace_place").get_value()
        use_call             = menu.get_option_by_name("use_call").get_value()

        self.dnumber = {}
        self.gen_handles = {}
        self.gen_keys = [[] for x in xrange(self.max_generations)]
        self.map = {}
        self.prev_gen_handles = {}

        if replace_date:
            empty_date = EMPTY_ENTRY
        else:
            empty_date = ""

        if replace_place:
            empty_place = EMPTY_ENTRY
        else:
            empty_place = ""

        self.__narrator = Narrator(self.database, self.verbose,
                                   use_call, self.use_fulldate,
                                   empty_date, empty_place,
                                   get_endnote_numbers=self.endnotes)

        self.bibli = Bibliography(Bibliography.MODE_PAGE)

    def endnotes(self, obj):
        if not obj or not self.inc_sources:
            return ""

        text = Enhanced_Endnotes.cite_source(self.bibli, obj)

        if text:
            text = "<super>" + text + "</super>"

        return text

    def write_report(self):
        """
        This function is called by the report system and writes the report.
        """

        self._create_person_list()

        self.doc.start_paragraph("DDR-Title")

        title = _(self.title)
        mark = IndexMark(title, INDEX_TYPE_TOC, 1)

        self.doc.write_text(title, mark)
        self.doc.end_paragraph()

        for generation in xrange(len(self.gen_keys)):
            if len(self.gen_keys[generation]):
                if self.pgbrk and generation > 0:
                    self.doc.page_break()

                self.doc.start_paragraph("DDR-Generation")

                text = _("Generation %d") % (generation+1)
                mark = IndexMark(text, INDEX_TYPE_TOC, 2)

                self.doc.write_text(text, mark)
                self.doc.end_paragraph()

                if self.desref:
                    self.prev_gen_handles = self.gen_handles.copy()
                    self.gen_handles.clear()

                for key in self.gen_keys[generation]:
                    person_handle = self.map[key]
                    self.gen_handles[person_handle] = key
                    self._write_person(key)

        if self.inc_sources:
            if self.pgbrkenotes:
                self.doc.page_break()

            Enhanced_Endnotes.write_endnotes(self.bibli, self.database,
                                             self.doc,
                                             printnotes=self.inc_srcnotes,
                                             links=self.links)

    # Filter for Record-style (Modified Register) numbering
    def _apply_mod_reg_filter(self, person_handle, cur_gen):
        self._apply_mod_reg_filter_aux(person_handle, len(self.map)+1, cur_gen)

        mod_reg_number = 1

        for generation in xrange(len(self.gen_keys)):
            for key in self.gen_keys[generation]:
                person_handle = self.map[key]
                self.dnumber[person_handle] = mod_reg_number

                mod_reg_number += 1

    def _apply_mod_reg_filter_aux(self, person_handle, index, cur_gen):
        if (not person_handle) or (cur_gen > self.max_generations) or \
           (person_handle in self.map.values()):
            return

        self.map[index] = person_handle

        if len(self.gen_keys) < cur_gen:
            self.gen_keys.append([index])
        else:
            self.gen_keys[cur_gen-1].append(index)

        person = self.database.get_person_from_handle(person_handle)

        for family_handle in person.get_family_handle_list():
            family = self.database.get_family_from_handle(family_handle)

            for child_ref in family.get_child_ref_list():
                ix = max(self.map)
                self._apply_mod_reg_filter_aux(child_ref.ref, ix + 1, self.gen_info[child_ref.ref])

    def _create_person_list(self):
        """
        Merge the results of each query in the filter to come up with a single list (in order)
        """

        people = self.filter.get_filter().apply(self.database)

        if len(people) > 1:
            rules = self.filter.get_filter().get_rules()

            if len(rules) > 1:
                from Filters import GenericFilter

                # This will wind up to be an intersection of all filter rules
                common_people = None

                # Find a person common to all the filter rules that we can use
                # to synchronize generation info on (the synchronized person is
                # generation 0, everybody else will be relative to that person)
                for rule in rules:
                    filter = GenericFilter()
                    filter.add_rule(rule)

                    if common_people:
                        common_people &= set(filter.apply(self.database))
                    else:
                        # First time through, start with the filtered list
                        common_people = set(filter.apply(self.database))

                if len(common_people) == 0:
                    raise ReportError(_("no common person found between filter rules to synchronize generations"))

                sync_handle = common_people.pop()
            else:
                # Only have the 1 rule, so everybody is common
                sync_handle = people[0]

            # Find connections between people so that we know what generation
            # they are in
            sync_person = self.database.get_person_from_handle(sync_handle)

            import Relationship
            relationship = Relationship.get_relationship_calculator()
        
            self.gen_info = {sync_handle: 0}
            min_gen = 0

            for handle in people:
                person = self.database.get_person_from_handle(handle)

                gen = 0
                prev = sync_person

                connections = find_deep_relations(self.database, None, sync_person, [], [], [handle])

                # Problem if somebody isn't connected
                if len(connections) == 0:
                    raise ReportError(_("can't determine generation for '%s'" % \
                        name_displayer.display(person)))

                for connection_handle in connections[0]:
                    person = self.database.get_person_from_handle(connection_handle)

                    if person != sync_person:
                        (str, d_orig, d_other) = relationship.get_one_relationship(self.database,
                                                                                   prev, person,
                                                                                   extra_info=True)

                        # Returning relationships with the distance to the common ancestor
                        # from first person (d_orig) and the second person (d_other).  The
                        # first preson is the 'prev'ious person as we travel down the list of
                        # people.  The second person is 'person'.
                        #
                        # If both distances are negative, the two people are not biologically
                        # related to each other and we assume they are both in the same generation.
                        #
                        # If both distances are positive, they are family members of the same
                        # generation.
                        #
                        # If the first person distance is 1 and the second person distance is 0,
                        # then the second person is an ancestor of the first person and is one generation
                        # up.
                        #
                        # If the first person distance is 0 and the second person distance is 1,
                        # then the second person is a descendant of the first person and is one generation
                        # down.
                        gen += (d_other - d_orig)

                        self.gen_info[handle] = gen

                        if gen < min_gen: min_gen = gen

                    prev = person

            offset = abs(min_gen)+1     # Want generations to start at 1, not 0

            for handle in self.gen_info:
                self.gen_info[handle] += offset
        else:
            # Setting up for a single individual
            self.gen_info = {people[0]: 1}

        # Now that we know which generation each person is in, we can merge the people
        # from all the rules in the filter
        for rule in self.filter.get_filter().get_rules():
            person = self.database.get_person_from_gramps_id(rule.list[0])

            if person != None:
                handle = person.get_handle()

                self._apply_mod_reg_filter(handle, self.gen_info[handle])
            else:
                raise ReportError(_("no person found with ID '%s'" % rule.list[0]))

    def _insert_image(self, database, doc, photo, w_cm=4.0, h_cm=4.0, alt=""):
        """
        Insert pictures of a person into the document.
        """

        object_handle = photo.get_reference_handle()
        media_object = database.get_object_from_handle(object_handle)
        mime_type = media_object.get_mime_type()

        if mime_type and mime_type.startswith("image"):
            filename = media_path_full(database, media_object.get_path())

            if os.path.exists(filename):
                doc.add_media_object(filename, "right", w_cm, h_cm, alt=alt,
                                     style_name="DDR-Caption", crop=photo.get_rectangle())
            else:
                from QuestionDialog import WarningDialog
                WarningDialog(_("Could not add photo to page"),
                                "%s: %s" % (filename, _('File does not exist')))

    def _write_children(self, family):
        """
        List the children for the given family.
        """

        if not family.get_child_ref_list():
            return

        mother_handle = family.get_mother_handle()
        if mother_handle:
            mother = self.database.get_person_from_handle(mother_handle)
            mother_name = name_displayer.display(mother)
        else:
            mother_name = _("unknown")

        father_handle = family.get_father_handle()
        if father_handle:
            father = self.database.get_person_from_handle(father_handle)
            father_name = name_displayer.display(father)
        else:
            father_name = _("unknown")

        self.doc.start_paragraph("DDR-ChildTitle")
        self.doc.write_text(
            _("Children of %(mother_name)s and %(father_name)s") %
                {'father_name': father_name,
                 'mother_name': mother_name
                })
        self.doc.end_paragraph()

        cnt = 1

        for child_ref in family.get_child_ref_list():
            child_handle = child_ref.ref
            child = self.database.get_person_from_handle(child_handle)
            child_name = name_displayer.display(child)
            child_mark = ReportUtils.get_person_mark(self.database, child)

            if self.desref and self.prev_gen_handles.get(child_handle):
                value = str(self.prev_gen_handles.get(child_handle))
                child_name += " [%s]" % value

            if self.inc_succession:
                prefix = " "

                for family_handle in child.get_family_handle_list():
                    family = self.database.get_family_from_handle(family_handle)

                    if family.get_child_ref_list():
                        prefix = "+ "
                        break
            else:
                prefix = ""

            if child_handle in self.dnumber:
                self.doc.start_paragraph("DDR-ChildList",
                    prefix
                    + str(self.dnumber[child_handle])
                    + " "
                    + ReportUtils.roman(cnt).lower()
                    + ".")
            else:
                self.doc.start_paragraph("DDR-ChildList",
                    prefix + ReportUtils.roman(cnt).lower() + ".")

            cnt += 1

            self.doc.write_text("%s. " % child_name, child_mark)
            self.__narrator.set_subject(child)
            self.doc.write_text_citation(self.__narrator.get_born_string() or
                                         self.__narrator.get_christened_string() or
                                         self.__narrator.get_baptised_string())
            self.doc.write_text_citation(self.__narrator.get_died_string() or
                                         self.__narrator.get_buried_string())
            self.doc.end_paragraph()

    def _write_event(self, event_ref):
        text = ""
        event = self.database.get_event_from_handle(event_ref.ref)

        # Don't write out private events if honoring the flag
        if self.honor_private and event.get_privacy() == True:
            return

        if self.use_fulldate:
            date = DateHandler.get_date(event)
        else:
            date = event.get_date_object().get_year()

        ph = event.get_place_handle()
        if ph:
            place = self.database.get_place_from_handle(ph).get_title()
        else:
            place = u""

        self.doc.start_paragraph('DDR-MoreDetails')
        event_name = str( event.get_type() )

        if date and place:
            text +=  _('%(date)s, %(place)s') % {
                'date' : date, 'place' : place
            }
        elif date:
            text += _('%(date)s') % {'date' : date}
        elif place:
            text += _('%(place)s') % { 'place' : place }

        if event.get_description():
            if text:
                text += ". "

            text += event.get_description()

        text += self.endnotes(event)

        if text:
            text += ". "

        text = _('%(event_name)s: %(event_text)s') % {
            'event_name' : _(event_name),
            'event_text' : text
        }

        self.doc.write_text_citation(text)

        if self.inc_attrs:
            text = ""
            attr_list = event.get_attribute_list()
            attr_list.extend(event_ref.get_attribute_list())

            for attr in attr_list:
                # Don't write out private attributes if honoring the flag
                if self.honor_private and attr.get_privacy() == True:
                    continue

                if text:
                    text += "; "

                text += _("%(type)s: %(value)s%(endnotes)s") % {
                    'type'     : attr.get_type(),
                    'value'    : attr.get_value(),
                    'endnotes' : self.endnotes(attr)
                }

            text = " " + text
            self.doc.write_text_citation(text)

        self.doc.end_paragraph()

        if self.inc_notes:
            # If the event or event reference has a note attached to it,
            # get the text and format it correctly
            notelist = event.get_note_list()
            notelist.extend(event_ref.get_note_list())

            for notehandle in notelist:
                note = self.database.get_note_from_handle(notehandle)
                self.doc.write_styled_note(note.get_styledtext(),
                                           note.get_format(),"DDR-MoreDetails",
                                           contains_html= note.get_type() == NoteType.HTML_CODE,
                                           links=self.links)

    def _write_family_events(self, family):
        """
        List the events for the given family.
        """

        if not family.get_event_ref_list():
            return

        mother_handle = family.get_mother_handle()
        if mother_handle:
            mother = self.database.get_person_from_handle(mother_handle)
            mother_name = name_displayer.display(mother)
        else:
            mother_name = _("unknown")

        father_handle = family.get_father_handle()
        if father_handle:
            father = self.database.get_person_from_handle(father_handle)
            father_name = name_displayer.display(father)
        else:
            father_name = _("unknown")

        first = True

        for event_ref in family.get_event_ref_list():
            if first:
                self.doc.start_paragraph('DDR-MoreHeader')
                self.doc.write_text(
                    _('More about %(mother_name)s and %(father_name)s:') % {
                      'mother_name' : mother_name,
                      'father_name' : father_name })
                self.doc.end_paragraph()

                first = False

            self._write_event(event_ref)

    def _write_marriage(self, person):
        """
        Output marriage sentence.
        """

        is_first = True

        for family_handle in person.get_family_handle_list():
            family = self.database.get_family_from_handle(family_handle)
            spouse_handle = ReportUtils.find_spouse(person, family)
            spouse = self.database.get_person_from_handle(spouse_handle)

            text = ""
            spouse_mark = ReportUtils.get_person_mark(self.database, spouse)

            text = self.__narrator.get_married_string(family, is_first)

            if text:
                self.doc.write_text_citation(text, spouse_mark)
                is_first = False

    def _write_mate(self, person, family):
        """
        Write information about the person's spouse/mate.
        """

        if person.get_gender() == Person.MALE:
            mate_handle = family.get_mother_handle()
        else:
            mate_handle = family.get_father_handle()

        if mate_handle:
            mate = self.database.get_person_from_handle(mate_handle)

            self.doc.start_paragraph("DDR-MoreHeader")
            name = name_displayer.display_formal(mate)
            mark = ReportUtils.get_person_mark(self.database, mate)

            if family.get_relationship() == FamilyRelType.MARRIED:
                self.doc.write_text(_("Spouse: %s") % name, mark)
            else:
                self.doc.write_text(_("Relationship with: %s") % name, mark)

            if name[-1:] != '.':
                self.doc.write_text(".")

            self.doc.write_text_citation(self.endnotes(mate))
            self.doc.end_paragraph()

            self._write_person_info(mate)

    def _write_path(self, person):
        path = []

        while True:
            # Person changes in the loop
            family_handle = person.get_main_parents_family_handle()

            if family_handle:
                family = self.database.get_family_from_handle(family_handle)
                mother_handle = family.get_mother_handle()
                father_handle = family.get_father_handle()

                if mother_handle and mother_handle in self.dnumber:
                    person = self.database.get_person_from_handle(mother_handle)
                    person_name = name_displayer.display_name(person.get_primary_name())
                    path.append(person_name)
                elif father_handle and father_handle in self.dnumber:
                    person = self.database.get_person_from_handle(father_handle)
                    person_name = name_displayer.display_name(person.get_primary_name())
                    path.append(person_name)
                else:
                    break
            else:
                break

        index = len(path)

        if index:
            self.doc.write_text("(")

        for name in path:
            if index == 1:
                self.doc.write_text(name + "-" + str(index) + ") ")
            else:
                self.doc.write_text(name + "-" + str(index) + "; ")

            index -= 1

    def _write_parents(self, person):
        family_handle = person.get_main_parents_family_handle()

        if family_handle:
            family = self.database.get_family_from_handle(family_handle)
            mother_handle = family.get_mother_handle()
            father_handle = family.get_father_handle()

            if mother_handle:
                mother = self.database.get_person_from_handle(mother_handle)
                mother_name = name_displayer.display_name(mother.get_primary_name())
                mother_mark = ReportUtils.get_person_mark(self.database, mother)
            else:
                mother_name = ""
                mother_mark = ""

            if father_handle:
                father = self.database.get_person_from_handle(father_handle)
                father_name = name_displayer.display_name(father.get_primary_name())
                father_mark = ReportUtils.get_person_mark(self.database, father)
            else:
                father_name = ""
                father_mark = ""

            text = self.__narrator.get_child_string(father_name, mother_name)

            if text:
                self.doc.write_text(text)

                if father_mark:
                    self.doc.write_text("", father_mark)

                if mother_mark:
                    self.doc.write_text("", mother_mark)

    def _write_person(self, key):
        """
        Output birth, death, parentage, marriage and notes information
        """

        person_handle = self.map[key]
        person = self.database.get_person_from_handle(person_handle)

        val = self.dnumber[person_handle]
        self.doc.start_paragraph("DDR-First-Entry", "%s." % val)

        name = name_displayer.display_formal(person)
        mark = ReportUtils.get_person_mark(self.database, person)

        self.doc.start_bold()
        self.doc.write_text(name, mark)
        if name[-1:] == '.':
            self.doc.write_text_citation("%s " % self.endnotes(person))
        else:
            self.doc.write_text_citation("%s. " % self.endnotes(person))
        self.doc.end_bold()

        if self.inc_paths:
            self._write_path(person)

        if self.omit_duplicate:
            # Check for duplicate record (result of distant cousins marrying)
            for dkey in sorted(self.map):
                if dkey >= key:
                    break

                if self.map[key] == self.map[dkey]:
                    self.doc.write_text(_("%(name)s is the same person as [%(id_str)s].") %
                                        { 'name' : '', 'id_str' : str(dkey) })
                    self.doc.end_paragraph()
                    return

        self.doc.end_paragraph()

        self._write_person_info(person)

        if self.listchildren or self.inc_events or self.inc_mates:
            for family_handle in person.get_family_handle_list():
                family = self.database.get_family_from_handle(family_handle)

                if self.inc_mates:
                    self._write_mate(person, family)

                if self.listchildren:
                    self._write_children(family)

                if self.inc_events:
                    self._write_family_events(family)

    def _write_person_info(self, person):
        name = name_displayer.display_formal(person)
        self.__narrator.set_subject(person)

        plist = person.get_media_list()
        if self.inc_photos and len(plist) > 0:
            if self.inc_captions:
                subst = SubstKeywords(self.database, person.get_handle())

                # Performing the following steps to get a single caption string:
                #
                # 1. Substitute any markers ($[a-z]) that need to get replaced
                # 2. Remove empty strings from the resultant list
                # 3. Convert the list to a single string of caption text items
                caption = subst.replace_and_clean(self.caption_fmt.get_value())
                caption = filter(lambda x: len(x)>0, caption)
                caption = '</text:p><text:p text:style-name="DDR-Caption">'.join(caption)

                self._insert_image(self.database, self.doc, plist[0], alt=caption)
            else:
                self._insert_image(self.database, self.doc, plist[0])

        self.doc.start_paragraph("DDR-Entry")

        if not self.verbose:
            self.__write_parents(person)

        text = self.__narrator.get_born_string()
        if text:
            self.doc.write_text_citation(text)

        text = self.__narrator.get_baptised_string()
        if text:
            self.doc.write_text_citation(text)

        text = self.__narrator.get_christened_string()
        if text:
            self.doc.write_text_citation(text)

        text = self.__narrator.get_died_string(self.calcage)
        if text:
            self.doc.write_text_citation(text)

        text = self.__narrator.get_buried_string()
        if text:
            self.doc.write_text_citation(text)

        if self.verbose:
            self._write_parents(person)

        self._write_marriage(person)
        self.doc.end_paragraph()

        notelist = person.get_note_list()
        if len(notelist) > 0 and self.inc_notes:
            self.doc.start_paragraph("DDR-NoteHeader")
            self.doc.write_text(_("Notes for %s") % name)
            self.doc.end_paragraph()

            for notehandle in notelist:
                note = self.database.get_note_from_handle(notehandle)
                self.doc.write_styled_note(note.get_styledtext(),
                                           note.get_format(),"DDR-Entry",
                                           contains_html= note.get_type() == NoteType.HTML_CODE,
                                           links=self.links)

        first = True

        if self.inc_names:
            for alt_name in person.get_alternate_names():
                if first:
                    self.doc.start_paragraph('DDR-MoreHeader')
                    self.doc.write_text(_('More about %(person_name)s:') % {
                        'person_name' : name
                    })
                    self.doc.end_paragraph()

                    first = False

                self.doc.start_paragraph('DDR-MoreDetails')
                atype = str( alt_name.get_type() )
                aname = alt_name.get_regular_name()
                self.doc.write_text_citation(_('%(name_kind)s: %(name)s%(endnotes)s') % {
                    'name_kind' : atype,
                    'name' : aname,
                    'endnotes' : self.endnotes(alt_name),
                })
                self.doc.end_paragraph()

        if self.inc_events:
            for event_ref in person.get_primary_event_ref_list():
                if first:
                    self.doc.start_paragraph('DDR-MoreHeader')
                    self.doc.write_text(_('More about %(person_name)s:') % {
                        'person_name' : name_displayer.display(person)
                    })
                    self.doc.end_paragraph()

                    first = False

                self._write_event(event_ref)

        if self.inc_addr:
            for addr in person.get_address_list():
                if first:
                    self.doc.start_paragraph('DDR-MoreHeader')
                    self.doc.write_text(_('More about %(person_name)s:') % {
                        'person_name' : name
                    })
                    self.doc.end_paragraph()

                    first = False

                self.doc.start_paragraph('DDR-MoreDetails')

                text = ReportUtils.get_address_str(addr)

                if self.use_fulldate:
                    date = DateHandler.get_date(addr)
                else:
                    date = addr.get_date_object().get_year()

                self.doc.write_text(_('Address: '))
                if date:
                    self.doc.write_text( '%s, ' % date )
                self.doc.write_text( text )
                self.doc.write_text_citation( self.endnotes(addr) )
                self.doc.end_paragraph()

        if self.inc_attrs:
            attrs = person.get_attribute_list()

            if first and attrs:
                self.doc.start_paragraph('DDR-MoreHeader')
                self.doc.write_text(_('More about %(person_name)s:') % {
                    'person_name' : name
                })
                self.doc.end_paragraph()

                first = False

            for attr in attrs:
                # Don't write out private attributes if honoring the flag
                if self.honor_private and attr.get_privacy() == True:
                    continue

                self.doc.start_paragraph('DDR-MoreDetails')
                text = _("%(type)s: %(value)s%(endnotes)s") % {
                    'type'     : attr.get_type(),
                    'value'    : attr.get_value(),
                    'endnotes' : self.endnotes(attr)
                }
                self.doc.write_text_citation( text )
                self.doc.end_paragraph()

#------------------------------------------------------------------------
#
# DescendantOptions
#
#------------------------------------------------------------------------
class DescendantOptions(MenuReportOptions):
    """
    Defines options and provides handling interface.
    """

    def __init__(self, name, database):
        self.database = database

        MenuReportOptions.__init__(self, name, database)

    def add_menu_options(self, menu):
        """
        Add options to the menu for the filtered descendant report.
        """

        category_name = _("Report Options")

        self.filter = FilterOption(_("Filter"), 0)
        self.filter.set_help(
            _("Determines what people are included in the report"))
        menu.add_option(category_name, "filter", self.filter)
        self.filter.connect("value-changed", self._filter_changed)

        from Filters import CustomFilters
        if len(CustomFilters.get_filters("Person")) == 0:
            raise ReportError(_("no custom filters found -- can only use custom filters with this report"))
        self.filter.set_filters(CustomFilters.get_filters("Person"))

        self.generations = NumberOption(_("Generations"), 10, 1, 100)
        self.generations.set_help(_("The number of generations to include in the " \
                               "report"))
        menu.add_option(category_name, "max_generations", self.generations)

        pgbrk = BooleanOption(_("Page break between generations"), False)
        pgbrk.set_help(
            _("Whether to start a new page after each generation."))
        menu.add_option(category_name, "pgbrk", pgbrk)

        pgbrkenotes = BooleanOption(_("Page break before end notes"),False)
        pgbrkenotes.set_help(
            _("Whether to start a new page before the end notes."))
        menu.add_option(category_name, "pgbrkenotes", pgbrkenotes)

        self.title = StringOption(_("Report title"), self.filter.get_filter().get_name())
        self.title.set_help(_("This text will be used as the report title"))
        menu.add_option(category_name, "title", self.title)

        category_name = _("Content")

        use_call = BooleanOption(_("Use callname for common name"), False)
        use_call.set_help(_("Whether to use the call name as the first name."))
        menu.add_option(category_name, "use_call", use_call)

        use_fulldate = BooleanOption(_("Use full dates instead of only the year"), True)
        use_fulldate.set_help(_("Whether to use full dates instead of just year."))
        menu.add_option(category_name, "use_fulldate", use_fulldate)

        listchildren = BooleanOption(_("List children"), True)
        listchildren.set_help(_("Whether to list children."))
        menu.add_option(category_name, "listchildren", listchildren)

        calcage = BooleanOption(_("Compute death age"),True)
        calcage.set_help(_("Whether to compute a person's age at death."))
        menu.add_option(category_name, "calcage", calcage)

        omit_duplicate = BooleanOption(_("Omit duplicate ancestors"), True)
        omit_duplicate.set_help(_("Whether to omit duplicate ancestors."))
        menu.add_option(category_name, "omit_duplicate", omit_duplicate)

        verbose = BooleanOption(_("Use complete sentences"), True)
        verbose.set_help(_("Whether to use complete sentences or succinct language."))
        menu.add_option(category_name, "verbose", verbose)

        desref = BooleanOption(_("Add descendant reference in child list"), True)
        desref.set_help(_("Whether to add descendant references in child list."))
        menu.add_option(category_name, "desref", desref)

        links = BooleanOption(_("Turn URLs into real links"), True)
        links.set_help(_("Turn any URLs starting with 'http', 'https', or 'mailto' into clickable links"))
        menu.add_option(category_name, "links", links)

        honor_private = BooleanOption(_("Do not include information marked private"), True)
        honor_private.set_help(_("Whether to honor the private flag setting"))
        menu.add_option(category_name, "honor_private", honor_private)

        category_name = _("Include")

        inc_notes = BooleanOption(_("Include notes"), True)
        inc_notes.set_help(_("Whether to include notes."))
        menu.add_option(category_name, "inc_notes", inc_notes)

        inc_attrs = BooleanOption(_("Include attributes"), False)
        inc_attrs.set_help(_("Whether to include attributes."))
        menu.add_option(category_name, "inc_attrs", inc_attrs)

        self.inc_photos = BooleanOption(_("Include Photo/Images from Gallery"), False)
        self.inc_photos.set_help(_("Whether to include images."))
        menu.add_option(category_name, "inc_photos", self.inc_photos)
        self.inc_photos.connect("value-changed", self._photo_changed)

        inc_names = BooleanOption(_("Include alternative names"), False)
        inc_names.set_help(_("Whether to include other names."))
        menu.add_option(category_name, "inc_names", inc_names)

        inc_events = BooleanOption(_("Include events"), False)
        inc_events.set_help(_("Whether to include events."))
        menu.add_option(category_name, "inc_events", inc_events)

        inc_addr = BooleanOption(_("Include addresses"), False)
        inc_addr.set_help(_("Whether to include addresses."))
        menu.add_option(category_name, "inc_addr", inc_addr)

        inc_sources = BooleanOption(_("Include sources"), False)
        inc_sources.set_help(_("Whether to include source references."))
        menu.add_option(category_name, "inc_sources", inc_sources)

        inc_srcnotes = BooleanOption(_("Include sources notes"), False)
        inc_srcnotes.set_help(_("Whether to include source notes in the "
                                "Endnotes section. Only works if 'Include sources' is selected."))
        menu.add_option(category_name, "inc_srcnotes", inc_srcnotes)

        inc_mates = BooleanOption(_("Include partners"), False)
        inc_mates.set_help(_("Whether to include detailed partner information."))
        menu.add_option(category_name, "inc_mates", inc_mates)

        inc_succession = BooleanOption(_("Include sign of succession ('+')" \
                                         " in child-list"), False)
        inc_succession.set_help(_("Whether to include a sign ('+') before the"
                                  " descendant number in the child-list to indicate"
                                  " a child has succession."))
        menu.add_option(category_name, "inc_succession", inc_succession)

        inc_paths = BooleanOption(_("Include path to start-person"), False)
        inc_paths.set_help(_("Whether to include the path of descendancy " \
                             "from the start-person to each descendant."))
        menu.add_option(category_name, "inc_paths", inc_paths)

        category_name = _("Caption")

        self.inc_captions = BooleanOption(_("Add caption under image from Gallery"), False)
        self.inc_captions.set_help(_("Whether to add captions underneath images."))
        menu.add_option(category_name, "inc_captions", self.inc_captions)

        self.caption_fmt = TextOption(_("Caption Format"), ["$n"] )
        self.caption_fmt.set_help(_("Display format for the caption."))
        menu.add_option(category_name, "caption", self.caption_fmt)

        category_name = _("Missing information")

        replace_place = BooleanOption(_("Replace missing places with ______"), False)
        replace_place.set_help(_("Whether to replace missing Places with blanks."))
        menu.add_option(category_name, "replace_place", replace_place)

        replace_date = BooleanOption(_("Replace missing dates with ______"), False)
        replace_date.set_help(_("Whether to replace missing Dates with blanks."))
        menu.add_option(category_name, "replace_date", replace_date)

    def make_default_style(self, default_style):
        """
        Make the default output style for the Filtered Descendant Report
        """

        font = FontStyle()
        font.set(face=FONT_SANS_SERIF, size=16, bold=1)
        para = ParagraphStyle()
        para.set_font(font)
        para.set_header_level(1)
        para.set_top_margin(0.25)
        para.set_bottom_margin(0.25)
        para.set_alignment(PARA_ALIGN_CENTER)
        para.set_description(_('The style used for the title of the page.'))
        default_style.add_paragraph_style("DDR-Title", para)

        font = FontStyle()
        font.set(face=FONT_SANS_SERIF, size=14, italic=1)
        para = ParagraphStyle()
        para.set_font(font)
        para.set_header_level(2)
        para.set_top_margin(0.25)
        para.set_bottom_margin(0.25)
        para.set_description(_('The style used for the generation header.'))
        default_style.add_paragraph_style("DDR-Generation", para)

        font = FontStyle()
        font.set(face=FONT_SANS_SERIF, size=10, italic=0, bold=1)
        para = ParagraphStyle()
        para.set_font(font)
        para.set_left_margin(1.5)   # in centimeters
        para.set_top_margin(0.25)
        para.set_bottom_margin(0.25)
        para.set_description(_('The style used for the children list title.'))
        default_style.add_paragraph_style("DDR-ChildTitle", para)

        font = FontStyle()
        font.set(size=10)
        para = ParagraphStyle()
        para.set_font(font)
        para.set_left_margin(1.524)
        para.set_top_margin(0.125)
        para.set_bottom_margin(0.125)
        para.set_description(_('The style used for the children list.'))
        default_style.add_paragraph_style("DDR-ChildList", para)

        font = FontStyle()
        font.set(face=FONT_SANS_SERIF, size=10, italic=0, bold=1)
        para = ParagraphStyle()
        para.set_font(font)
        para.set(first_indent=0.0, lmargin=1.5)
        para.set_top_margin(0.25)
        para.set_bottom_margin(0.25)
        default_style.add_paragraph_style("DDR-NoteHeader", para)

        para = ParagraphStyle()
        para.set(lmargin=1.5)
        para.set_top_margin(0.25)
        para.set_bottom_margin(0.25)
        para.set_description(_('The basic style used for the text display.'))
        default_style.add_paragraph_style("DDR-Entry", para)

        para = ParagraphStyle()
        para.set(first_indent=-1.5, lmargin=1.5)
        para.set_top_margin(0.25)
        para.set_bottom_margin(0.25)
        para.set_description(_('The style used for the first personal entry.'))
        default_style.add_paragraph_style("DDR-First-Entry", para)

        font = FontStyle()
        font.set(size=10, face=FONT_SANS_SERIF, bold=1)
        para = ParagraphStyle()
        para.set_font(font)
        para.set(first_indent=0.0, lmargin=1.5)
        para.set_top_margin(0.25)
        para.set_bottom_margin(0.25)
        para.set_description(_('The style used for the More About header.'))
        default_style.add_paragraph_style("DDR-MoreHeader", para)

        font = FontStyle()
        font.set(face=FONT_SERIF, size=10)
        para = ParagraphStyle()
        para.set_font(font)
        para.set(first_indent=0.0, lmargin=1.5)
        para.set_top_margin(0.25)
        para.set_bottom_margin(0.25)
        para.set_description(_('The style used for additional detail data.'))
        default_style.add_paragraph_style("DDR-MoreDetails", para)

        font = FontStyle()
        font.set(face=FONT_SERIF, size=10)
        para = ParagraphStyle()
        para.set_font(font)
        para.set_alignment(PARA_ALIGN_CENTER)
        para.set_description(_('The style used for photo captions.'))
        default_style.add_paragraph_style("DDR-Caption", para)

        Enhanced_Endnotes.add_endnote_styles(default_style)

    def _filter_changed(self):
        try:
            # Update the title to match the name of the filter
            self.title.set_value(self.filter.get_filter().get_name())
        except AttributeError:
            pass

    def _photo_changed(self):
        self.inc_captions.set_available(self.inc_photos.get_value())
        self.caption_fmt.set_available(self.inc_photos.get_value())

